import os, json, time
import logging
from collections import defaultdict
from jsonpath_rw import parse

from rltk.tokenizer.digCrfTokenizer.crf_tokenizer import ngramTokenizer
from rltk.configuration import Configuration
import rltk.utils as utils

from datasketch import MinHash, MinHashLSH

class MinHashLSHRecordDeduplication(object):
    """
      Base Class to build Minhash LSH indexer for a single database. This class has methods to
      a) processes records in batches
      b) construct minhash lsh based signatures
      c) write minhash lsh indexes to disk
    """
    BATCH_SIZE = 20000
    NUM_PERMUTATIONS = 128
    THRESHOLD = 0.9
    INDEX_THRESHOLD = 1000
    BANDS = 0
    ROWS = 0
    
    _block_index = 0

    def __init__(self, **kwargs):
        """
          Initialiser to setup variables

          Args:
          **kwargs: Arbitrary keyword arguments

          Returns:
            None
        """
        self._check_args(kwargs)
        self._kwargs = kwargs
        
        # MinHash LSH specific params
        if 'bands_rows' in self._kwargs:
            (self._bands, self._rows) = self._kwargs['bands_rows']
        else:
            (self._bands, self._rows) = (MinHashLSHRecordDeduplication.BANDS, MinHashLSHRecordDeduplication.ROWS)
        if 'num_perm' in self._kwargs:
            self._num_perm = int(self._kwargs['num_perm'])
        else:
            self._num_perm = MinHashLSHRecordDeduplication.NUM_PERMUTATIONS
        if 'threshold' in self._kwargs:
            self._threshold = float(self._kwargs['threshold'])
        else:
            self._threshold = MinHashLSHRecordDeduplication.THRESHOLD
        
        # Common params
        self.output_file_path = self._kwargs['output_file_path']
        self.value_path = self._kwargs['value_path1']
        self._discarded_indexes = set()
        self._file_iter = self._kwargs['iter1']
        
        if 'batch_size' in self._kwargs:
            self._batch_size = int(self._kwargs['batch_size'])
        else:
            self._batch_size = MinHashLSHRecordDeduplication.BATCH_SIZE

        self.erase_file_content(self.output_file_path)
        self.output_flag = False
        self.nt = ngramTokenizer()
        self._logger = logging.getLogger(Configuration.LOGGER_NAME)

    def erase_file_content(self, file_path):
        """
          This erases file contents if the file exists.

          Args:
            file_path(string) : file path

          Returns:
            None
        """
        if os.path.isfile(file_path) and os.stat(file_path).st_size > 0:
            open(self.output_file_path, 'w').close()

    def _check_args(self, kwargs):
        """
          Helper method to check if
          a) necessary arguments are passed
          b) Argument values are of correct (value / type)

          Args:
            kwargs (dict): Dictionary of argument key value pairs

          Returns:
            None
        """
        error_flag = False

        # Common params
        if 'output_file_path' not in kwargs:
            error_flag = True
            msg = 'Missing output file path argument - output_file_path'
        if 'value_path1' not in kwargs:
            error_flag = True
            msg = 'Missing blocking value path argument- value_path'
        if 'iter1' not in kwargs:
            error_flag = True
            msg = 'Missing file iterator argument - iter1'

        if error_flag:
            raise ValueError(msg)

    def _write_indexer(self, indexer):
        """
          Writes indexer back to disk
          1. Read from indexer disk file if it exists
          2. update indexer by adding newly added block candidates and write to temp file
          Finally write back to original file and delete temp file

          Args:
            indexer (dict): Dictionary of block ID and corresponding record ids

          Returns:
            None
            Has a side effect of writing to output file
        """
        self.output_file_path = self._kwargs['output_file_path']

        if self.output_flag:
            filename, file_extension = os.path.splitext(self.output_file_path)
            temp_file = os.path.join(self.output_file_path + '.temp')
            temp_fptr = open(temp_file, 'w')
            with open(self.output_file_path, 'r') as of:
                for line in of:
                    jline = json.loads(line)
                    for k in jline.keys():
                        # merge qgram indexes if they already exist
                        if k in indexer:
                            s = set(jline[k])
                            s |= set(indexer[k])
                            # check if length of qgram indexes threshold size
                            if len(s) > self.INDEX_THRESHOLD:
                                self._discarded_indexes.add(k)
                            if k not in self._discarded_indexes:
                                json.dump({k: list(s)}, temp_fptr)
                                temp_fptr.write('\n')

                            # we have already seen key k
                            del indexer[k]
                        else:
                            json.dump(jline, temp_fptr)
                            temp_fptr.write('\n')

            # Write remaining indexes to temp file
            for k, v in indexer.items():
                if len(v) < self.INDEX_THRESHOLD:
                    json.dump({k: v}, temp_fptr)
                    temp_fptr.write('\n')
                else:
                    self._discarded_indexes.add(k)
            temp_fptr.close()

            # write back temp file to output file line by line (less memory)
            op_fptr = open(self.output_file_path, 'w')
            with open(temp_file, 'r') as tf:
                for line in tf:
                    jline = json.loads(line)
                    json.dump(jline, op_fptr)
                    op_fptr.write('\n')
            op_fptr.close()
            # delete temp file
            os.remove(temp_file)
        else:
            # Writing to output file for first time
            with open(self.output_file_path, 'w') as of:
                for k, v in indexer.items():
                    if len(v) < self.INDEX_THRESHOLD:
                        json.dump({k: v}, of)
                        of.write('\n')
                    else:
                        self._discarded_indexes.add(k)
            self.output_flag = True

    def _index_records(self, records):
        """
          Constructs Minhash LSH buckets for a given set of records

          Args:
            records (dict) : dict of (record_id -> record_value)

          Returns:
            None
        """
        indexer = defaultdict(list)
        
        # Create minhashes
        minhashes = {}
        for rid in records:
            m = MinHash(num_perm=self._num_perm)
            for d in records[rid]:
                qgrams = set(self.nt.basic(d, 2))
                for gram in qgrams:
                    m.update(gram.encode('utf-8'))
            minhashes[rid] = m
        
        # Create LSH instance and add min hashes
        if self._bands == MinHashLSHRecordDeduplication.BANDS and self._rows == MinHashLSHRecordDeduplication.ROWS:
            lsh = MinHashLSH(threshold=self._threshold,num_perm=self._num_perm)
        else:
            lsh = MinHashLSH(num_perm=self._num_perm, params=(self._bands, self._rows))
            
        max_blocks = []
        for rid in records:
            lsh.insert(rid, minhashes[rid])
            max_blocks.append(rid)
        
        # Generate blocks
        while(len(max_blocks)>0):
            key = max_blocks[0]
            bucket = lsh.query(minhashes[key])
            for rid in bucket:
                if rid in max_blocks:
                    max_blocks.remove(rid)
                indexer["b"+str(self._block_index)].append(rid)
            self._block_index += 1
        
        self._write_indexer(indexer)

    def build_index(self):
        """
          Builds MinHash LSH blocking indexer for single database. It processes records in batches of BATCH_SIZE.

          Args:
            None

          Returns:
            None
            Has a side effect of building MinHash LSH indexer and writing indexer to disk.
        """
        records = {}
        run_count = 0
        run_iteration = 1
        parse_dict = {}
        for k in self.value_path:
            parse_dict[k] = parse(k)
        s = time.time()
        for rid, json_data in self._file_iter:
            extracted_data = utils.extract(json_data, self.value_path, parse_dict)
            # Reset run_count when we hit BATCH_SIZE
            if run_count >= self._batch_size:
                self._index_records(records)
                msg = "Finished indexing {val} records. Time = {time}".format(val=run_count * run_iteration,
                                                                              time=(time.time() - s))
                self._logger.info('{0} {1}'.format("[minhash-lsh-blocking]", msg))

                run_iteration += 1
                records = {}
                run_count = 0

            records[rid] = set(extracted_data.values()[0])
            run_count += 1

        # Index the final remaining records
        self._index_records(records)

    def get_inverted_index(self):
        """
          Builds inverted index from the index file

          Args:
            None

          Returns:
            dict: {record_id: [minhash-lsh-bucketid]}
        """
        inverted_index = defaultdict(list)
        with open(self.output_file_path, 'r') as ip:
            for line in ip:
                jline = json.loads(line)
                for k in jline.keys():
                    id_list = jline[k]
                    for id_v in id_list:
                        inverted_index[id_v].append(k)

        return inverted_index

    def get_index(self):
        """
          Builds index from index file

          Args:
            None

          Returns:
            dict: {minhash-lsh-bucketid: [record_ids]}
        """
        index = defaultdict(list)
        with open(self.output_file_path, 'r') as ip:
            for line in ip:
                jline = json.loads(line)
                for k in jline.keys():
                    index[k].extend(jline[k])

        return index

def minhash_lsh_indexing(**kwargs):
    """
      Base interface to construct Minhash LSH indexes for databases.

      Args:
        **kwargs: Arbitrary keyword arguments

      Returns:
        None
        Has side effect of writing Minhash LSH indexes to file
    """
    if 'iter2' in kwargs:
        q = MinHashLshRecordLinkage(**kwargs)
        q.build_index()
    else:
        q = MinHashLSHRecordDeduplication(**kwargs)
        q.build_index()



from rltk.record import Record, AutoGeneratedRecord,\
    cached_property, generate_record_property_cache, validate_record, remove_raw_object
from rltk.dataset import Dataset
from rltk.io import *
from rltk.similarity import *
from rltk.blocking import *
from rltk.tokenizer import *
from rltk.evaluation import *
from rltk.utils import get_record_pairs
from rltk.parallel_processor import ParallelProcessor
from rltk.map_reduce import MapReduce
import rltk.cli
import rltk.remote



from threading import Semaphore
from typing import Callable

from rltk.remote.remote import Remote


class Task(object):
    """
    Remote task. It has similar API to :meth:`rltk.ParallelProcessor`. 
    But do not use :meth:`rltk.ParallelProcessor` if this module is used. If you still want multiprocessing, 
    please give each worker more processes.
    
    Args:
        remote (Remote): Remote object.
        input_handler (Callable): Input handler.
        output_handler (Callable): Output handler. It accepts same number of arguments to `input_handler` 's return values.
        chunk_size (int, optional): Size of the each data chunk. Defaults to 1000.
        max_queue_size (int, optional): How many chunks can be in the queue. Defaults to 10.
    """

    def __init__(self, remote: Remote, input_handler: Callable, output_handler: Callable,
                 chunk_size: int = 1000, max_queue_size: int = 10):
        self.remote = remote
        self.input_handler = input_handler
        self.output_handler = output_handler

        self.chunk_data = []  # buffer
        self.chunk_size = chunk_size  # buffer size
        self.future_semaphore = Semaphore(value=max_queue_size)  # max num of un-return futures
        self.all_futures = set([])  # all un-return future objects
        self.done = False

    def start(self):
        """
        Start listening.
        """
        pass

    @staticmethod
    def _parse_input(input_handler, data):
        return [input_handler(*args, **kwargs) for args, kwargs in data]

    def _parse_output(self, future):
        if future.done():
            for r in future.result():
                if not isinstance(r, tuple):
                    r = (r,)
                self.output_handler(*r)

        # release resources no matter what condition that future gets
        self.all_futures.remove(future)
        self.future_semaphore.release()

    def compute(self, *args, **kwargs):
        """
        Add data to compute.
        """
        if self.done:
            return

        if len(self.chunk_data) < self.chunk_size:
            self.chunk_data.append(([*args], {**kwargs}))
        if len(self.chunk_data) == self.chunk_size:
            self._submit()

    def _submit(self):
        if len(self.chunk_data) == 0:
            return

        self.future_semaphore.acquire()

        # scatter input data (scatter first if data is large)
        data_future = self.remote.scatter(self.chunk_data)

        # input and output must be staticmethod, create wrappers to bypass restriction
        future = self.remote.submit(Task._parse_input, self.input_handler, data_future)
        Task._parse_output_wrapper = lambda ft: Task._parse_output(self, ft)

        # add listener
        future.add_done_callback(Task._parse_output_wrapper)
        self.all_futures.add(future)

        self.chunk_data = []

    def task_done(self):
        """
        Indicate that all resources which need to compute are added.
        """
        self.done = True
        self._submit()  # force flush buffer

    def join(self):
        """
        Block until all tasks are done.
        """
        while len(self.all_futures) != 0:
            pass



import io


class Writer(object):
    """
    Writer.
    """
    def __init__(self):
        pass

    def write(self):
        """
        Write content.
        """
        raise NotImplementedError

    def __del__(self):
        """
        Same to :meth:`close`.
        """
        self.close()

    def close(self):
        """
        Close handler.
        """
        pass



import rltk.utils as utils


def ngram_distance(s0, s1, n=2):
    """
    N-Gram Distance as defined by Kondrak, "N-Gram Similarity and Distance" String Processing and Information Retrieval, Lecture Notes in Computer Science Volume 3772, 2005, pp 115-126.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: NGram Distance.

    Examples:
        >>> rltk.ngram_distance('ABCD', 'ABTUIO')
        0.5833
    """

    utils.check_for_none(s0, s1)
    utils.check_for_type(str, s0, s1)

    n1, n2 = len(s0), len(s1)
    special = "\n"

    if (n1 == 0 or n2 == 0):
        return 1

    if (s0 == s1):
        return 0

    cost = 0
    if (n1 < n or n2 < n):
        return 1

    # Adding special chars (n-1) to s0
    sa = special * (n - 1) + s0

    s2_j = [None] * n  # jth n-gram of s2
    d = [0] * (n1 + 1)  # cost array, horizontally
    p = [0] * (n1 + 1)  # 'previous' cost array, horizontally

    for i in range(n1 + 1):
        p[i] = i

    for j in range(1, n2 + 1):
        # Construct s2_j n-gram
        if (j < n):
            for ti in range(n - j):
                s2_j[ti] = special

            for ti in range(n - j, n):
                s2_j[ti] = s1[ti - (n - j)]

        else:
            s2_j = list(s1[j - n: j])

        d[0] = j

        for i in range(1, n1 + 1):
            cost = 0
            tn = n
            # Compare sa to s2_j
            for ni in range(n):
                if sa[i - 1 + ni] != s2_j[ni]:
                    cost += 1
                elif sa[i - 1 + ni] == special:
                    tn -= 1

            ec = float(cost) / tn
            # minimum of cell to the left+1, to the top+1,
            # diagonally left and up +cost
            d[i] = min(d[i - 1] + 1, p[i] + 1, p[i - 1] + ec)

        d2 = p
        p = d
        d = d2
    return float(p[n1]) / max(n2, n1)


def ngram_similarity(s0, s1, n=2):
    """
    N-Gram Similarity as defined by Kondrak, "N-Gram Similarity and Distance" String Processing and Information Retrieval, Lecture Notes in Computer Science Volume 3772, 2005, pp 115-126.

    Args:
        s1 (str): Sequence 1.
        s2 (str): Sequence 2.

    Returns:
        float: NGram Similarity.

    Examples:
        >>> rltk.ngram_similarity('ABCD', 'ABTUIO')
        0.4166666666666667
    """

    utils.check_for_none(s0, s1)
    utils.check_for_type(str, s0, s1)

    n1, n2 = len(s0), len(s1)
    special = "\n"

    if (n1 == 0 or n2 == 0):
        return 0

    if (s0 == s1):
        return 1

    cost = 0
    if (n1 < n or n2 < n):
        return 0

    # Adding special chars (n-1) to s0
    sa = special * (n - 1) + s0

    s2_j = [None] * n  # jth n-gram of s2
    d = [0] * (n1 + 1)  # cost array, horizontally
    p = [0] * (n1 + 1)  # 'previous' cost array, horizontally

    for i in range(n1 + 1):
        p[i] = 0

    for j in range(1, n2 + 1):
        # Construct s2_j n-gram
        if (j < n):
            for ti in range(n - j):
                s2_j[ti] = special

            for ti in range(n - j, n):
                s2_j[ti] = s1[ti - (n - j)]

        else:
            s2_j = list(s1[j - n: j])

        d[0] = 0

        for i in range(1, n1 + 1):
            cost = 0
            tn = n
            # Compare sa to s2_j
            for ni in range(n):
                if sa[i - 1 + ni] == s2_j[ni] and sa[i - 1 + ni] != "\n":
                    cost += 1
                elif sa[i - 1 + ni] == special:
                    tn -= 1

            ec = float(cost) / tn
            # minimum of cell to the left+1, to the top+1,
            # diagonally left and up +cost
            d[i] = max(d[i - 1], p[i], p[i - 1] + ec)

        d2 = p
        p = d
        d = d2
    return float(p[n1]) / max(n2, n1)



import pandas as pd
import csv
import json
import io

from rltk.io.reader import *


arr = [{'1': 'A', '2': 'B'}, {'1': 'a', '2': 'b'}]


def test_array_reader():
    for idx, obj in enumerate(ArrayReader(arr)):
        assert obj == arr[idx]


def test_dataframe_reader():
    df = pd.DataFrame(arr)
    for idx, obj in enumerate(DataFrameReader(df)):
        assert obj == arr[idx]


def test_dataframe_reader_keep_index():
    df = pd.DataFrame(arr)
    for idx, obj in enumerate(DataFrameReader(df, True)):
        assert obj == dict(**arr[idx], dataframe_default_index=idx)


def test_csv_reader():
    f = io.StringIO()

    writer = csv.DictWriter(f, fieldnames=['1', '2'])
    writer.writeheader()
    for a in arr:
        writer.writerow(a)

    for idx, obj in enumerate(CSVReader(f)):
        assert obj == arr[idx]

    f.close()


def test_jsonlines_reader():
    f = io.StringIO()

    for a in arr:
        f.write(json.dumps(a) + '\n')

    for idx, obj in enumerate(JsonLinesReader(f)):
        assert obj == arr[idx]

    f.close()


# distance
from rltk.similarity.distance import euclidean_distance, manhattan_distance

# normal
from rltk.similarity.equal import string_equal, number_equal
from rltk.similarity.hamming import hamming_distance, hamming_similarity, normalized_hamming_distance
from rltk.similarity.dice import dice_similarity
from rltk.similarity.levenshtein import levenshtein_distance, levenshtein_similarity, \
    normalized_levenshtein_distance, damerau_levenshtein_distance, optimal_string_alignment_distance
from rltk.similarity.needleman import needleman_wunsch_score, needleman_wunsch_similarity
from rltk.similarity.jaro import jaro_winkler_distance, jaro_winkler_similarity, jaro_distance
from rltk.similarity.jaccard import jaccard_index_similarity, jaccard_index_distance
from rltk.similarity.cosine import cosine_similarity, string_cosine_similarity
from rltk.similarity.tf_idf import tf_idf_similarity, compute_idf, compute_tf, tf_idf_cosine_similarity, TF_IDF
from rltk.similarity.lcs import longest_common_subsequence_distance, metric_longest_common_subsequence
from rltk.similarity.ngram import ngram_distance, ngram_similarity
from rltk.similarity.qgram import qgram_distance, qgram_similarity

# # hybrid
from rltk.similarity.hybrid import hybrid_jaccard_similarity, monge_elkan_similarity, symmetric_monge_elkan_similarity

# # phonetic
from rltk.similarity.soundex import soundex_similarity, soundex
from rltk.similarity.metaphone import metaphone_similarity, metaphone
from rltk.similarity.nysiis import nysiis_similarity, nysiis



import json
import hashlib
import operator

from rltk.blocking.block import Block
from rltk.io.adapter.key_set_adapter import KeySetAdapter
from rltk.io.adapter.memory_key_set_adapter import MemoryKeySetAdapter


class BlockingHelper(object):
    """
    Blocking Helper.
    """

    @staticmethod
    def encode_inverted_index_key(dataset_id, record_id):
        return json.dumps({'d': dataset_id, 'r': record_id}, sort_keys=True)

    @staticmethod
    def decode_inverted_index_key(key):
        key = json.loads(key)
        return key['d'], key['r']

    @staticmethod
    def generate_inverted_indices(block: Block, ks_adapter: KeySetAdapter = None):
        """
        Generate inverted indices of block.
        
        Args:
            block (Block): Original block.
            ks_adapter (KeySetAdapter): Where the inverted indices store.
            
        Returns:
            KeySetAdapter:
        """
        if not ks_adapter:
            ks_adapter = MemoryKeySetAdapter()
        for block_id, dataset_id, record_id in block:
            ks_adapter.add(BlockingHelper.encode_inverted_index_key(dataset_id, record_id), block_id)
        return ks_adapter

    @staticmethod
    def _block_operations(operator_, left_block, right_block, right_inverted, output_block):
        operation = None
        if operator_ == 'union':
            operation = operator.or_  # lambda a, b: a | b
        elif operator_ == 'intersect':
            operation = operator.and_  # lambda a, b: a & b

        for left_block_id, left_data in left_block.key_set_adapter:
            for left_dataset_id, left_record_id in left_data:
                key = BlockingHelper.encode_inverted_index_key(left_dataset_id, left_record_id)
                right_block_ids = right_inverted.get(key)
                if right_block_ids:
                    for right_block_id in right_block_ids:
                        new_block_data = operation(left_data, right_block.get(right_block_id))
                        new_block_id = hashlib \
                            .sha1(''.join(sorted(['{},{}'.format(ds, r) for ds, r in new_block_data]))
                                  .encode('utf-8')).hexdigest()
                        output_block.key_set_adapter.set(new_block_id, new_block_data)

    @staticmethod
    def union(block1, inverted1, block2, inverted2, block3=None):
        """
        Union of two blocks.
        
        Args:
            block1 (Block): Block 1.
            inverted1 (KeySetAdapter): Inverted indices of block 1.
            block2 (Block): Block2.
            inverted2 (KeySetAdapter): Inverted indices of block 2.
            block3 (Block, optional): Unioned block. If None, a Block object will be created. Defaults to None. 
        
        Returns:
            Block:
        """
        block3 = block3 or Block()

        BlockingHelper._block_operations('union', block1, block2, inverted2, block3)
        BlockingHelper._block_operations('union', block2, block1, inverted1, block3)
        return block3

    @staticmethod
    def intersect(block1, inverted1, block2, inverted2, block3=None):
        """
        Intersection of two blocks.
        
        Args:
            block1 (Block): Block 1.
            inverted1 (KeySetAdapter): Inverted indices of block 1.
            block2 (Block): Block2.
            inverted2 (KeySetAdapter): Inverted indices of block 2.
            block3 (Block, optional): Intersected block. If None, a Block object will be created. Defaults to None. 
        
        Returns:
            Block:
        """
        block3 = block3 or Block()

        BlockingHelper._block_operations('intersect', block1, block2, inverted2, block3)
        BlockingHelper._block_operations('intersect', block2, block1, inverted1, block3)
        return block3



import pandas as pd

from rltk.io.reader import Reader


class DataFrameReader(Reader):
    """
    Pandas DataFrame Reader.
    
    Args:
        df (pandas.DataFrame): DataFrame.
        keep_dataframe_default_index (bool, optional): if True, add a key "dataframe_default_index" holding the \
                                                       original index in df. [default value: False]

    """

    def __init__(self, df: pd.DataFrame, keep_dataframe_default_index: bool=False):
        self._df = df
        self._keep_dataframe_default_index = keep_dataframe_default_index

    def __next__(self):
        if self._keep_dataframe_default_index:
            for i, item in self._df.iterrows():
                yield dict(item.to_dict(), dataframe_default_index=i)
        else:
            for _, item in self._df.iterrows():
                yield item.to_dict()


class Serializer(object):
    """
    Serialize and deserialize object. This is the super class.
    """

    def loads(self, obj):
        """
        Load a serialized object.
         
        Args:
            obj (obj): For most of times, it's byte or string.
            
        Returns:
            obj: Python object.
        """
        raise NotImplementedError

    def dumps(self, obj):
        """
        Serialize the given object.
        
        Args:
            obj (obj): Python object.
            
        Returns:
            obj: Serialized object.
        """
        raise NotImplementedError



from rltk.io.reader import Reader
from rltk.io.adapter import KeyValueAdapter, MemoryKeyValueAdapter
from rltk.record import Record, generate_record_property_cache, get_property_names
from rltk.parallel_processor import ParallelProcessor

import pandas as pd
from typing import Callable


class Dataset(object):
    """
    A set of records.
    
    Args:
        reader (Reader, optional): Input reader.
        record_class (type(Record), optional): Concrete class of :meth:`Record`.
        adapter (KeyValueAdapter, optional): Specify where to store indexed data. Defaults to :meth:`MemoryKeyValueAdapter`.
        size (int, optional): Same as `size` in :meth:`add_records` .
        pp_num_of_processor (int, optional): Same as `pp_num_of_processor` in :meth:`add_records` .
        pp_max_size_per_input_queue (int, optional): Same as `pp_max_size_per_input_queue` in :meth:`add_records` .
        sampling_function (callable, optional): Sampling function, `raw_object` is the only parameter. \
                                                If it returns True, record instance will be created.
    Note:
        - Set `reader`, `record_class` and `adapter` if new a :meth:`Dataset` needs to be generated.
        - If :meth:`Dataset` is already generated and stored in a permanent adapter, only adapter needs to be provided.
        - `reader` can also be added later, but `record_class` needs to be provided.
        
    """

    _METADATA_KEY = '__rltk_reserved_metadata'

    def __init__(self, reader: Reader = None, record_class: type(Record) = None, adapter: KeyValueAdapter = None,
                 size: int = None, sampling_function: Callable = None,
                 pp_num_of_processor: int = 0, pp_max_size_per_input_queue: int = 200):
        # load adapter and metadata if it's there
        self._adapter = adapter or MemoryKeyValueAdapter()
        metadata = self._adapter.get(self._METADATA_KEY)
        if metadata:
            self.id = metadata.get('id')
            self._record_class = metadata.get('record_class')

        # add / update
        self._sampling_function = sampling_function
        if record_class:
            if hasattr(self, '_record_class'):
                if self._record_class != record_class:
                    raise ValueError('Different record class already exists in current adapter')
            else:
                # add metadata
                self._record_class = record_class
                self.id = self._record_class.__name__
                metadata = {'id': self.id, 'record_class': self._record_class}
                self._adapter.set(self._METADATA_KEY, metadata)
            # add data
            if reader:
                self.add_records(reader, size, pp_num_of_processor, pp_max_size_per_input_queue)

    def add_records(self, reader: Reader, size: int = None,
                    pp_num_of_processor: int = 0, pp_max_size_per_input_queue: int = 200):
        """
        Add `records` to :meth:`Dataset` from `reader` .
        
        Args:
            reader (Reader, optional): Input reader.
            size (int, optional): Maximum size of records will be in :meth:`Dataset`, \
                            defaults to None, can be used when debugging.
            pp_num_of_processor (int, optional): Same as `num_of_processor` in :meth:`ParallelProcessor`. \ 
                            If non-zero is given and adapter is `parallel_safe`, \
                            this method will run in parallel mode. Defaults to 0.
            pp_max_size_per_input_queue (int, optional): Same as `max_size_per_input_queue` in \
                                    :meth:`ParallelProcessor`. Defaults to 200.
        """

        def generate(_raw_object):
            if not self._sampling_function or self._sampling_function(_raw_object):
                record_instance = self._record_class(_raw_object)
                generate_record_property_cache(record_instance)
                self._adapter.set(record_instance.id, record_instance)

        if not self._record_class:
            raise ValueError('Record class is not provided.')
        curr_size = 0

        # serial
        if pp_num_of_processor == 0 or not self._adapter.parallel_safe:
            for raw_object in reader:
                if not self._sampling_function or self._sampling_function(raw_object):
                    generate(raw_object)
                    curr_size += 1
                    if size and curr_size >= size:
                        break
        # parallel
        else:
            pp = ParallelProcessor(input_handler=generate, num_of_processor=pp_num_of_processor,
                                   max_size_per_input_queue=pp_max_size_per_input_queue)
            pp.start()
            for raw_object in reader:
                if not self._sampling_function or self._sampling_function(raw_object):
                    pp.compute(raw_object)
                    curr_size += 1
                    if size and curr_size >= size:
                        break
            pp.task_done()
            pp.join()

    def set_record(self, record: Record):
        """
        Setter of a record.
        
        Args:
            record (Record): Record object.
        """
        return self._adapter.set(record.id, record)

    def get_record(self, record_id):
        """
        Getter of a record.
        
        Args:
            record_id (str): Record id.
            
        Returns:
            Record: Record object.
        """
        return self._adapter.get(record_id)

    def generate_dataframe(self, size: int = None, **kwargs):
        """
        Generate Pandas Dataframe
        
        Args:
            size (int, optional): How many records should be used to generate `pandas.Dataframe`. 
                                None means all. Defaults to None.
        
        Returns:
            pandas.Dataframe
        """
        table = []
        columns = None
        count = 0

        # construct table
        for r in self:
            if size and count >= size:
                break
            count += 1

            # generate columns based on first record
            if not columns:
                columns = get_property_names(r.__class__)

            # get data
            row_data = []
            for prop_name in columns:
                row_data.append(getattr(r, prop_name))

            # append data
            table.append(row_data)

        return pd.DataFrame(table, columns=columns, **kwargs)

    def head(self, n: int = 10):
        """
        Iterate on first n records.

        Args:
            n (int, optional): Number of records. Default to 10.

        Returns:
            iter: Record
        """
        for idx, r in enumerate(self.__next__()):
            if idx >= n:
                break
            yield r

    def filter(self, condition: Callable, n: int = 10):
        """
        Iterate on filtered n records.

        Args:
            condition (Callable): `function(r: Record) -> bool`. If the return is True, record will be selected.
            n (int, optional): Number of records. Default to 10.

        Returns:
            iter: Record
        """
        counter = 0
        for r in self.__next__():
            if counter >= n:
                break
            if condition(r):
                counter += 1
                yield r

    def __iter__(self):
        """
        Same as :meth:`__next__`
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator
        
        Returns:
            iter: Record
        """
        for k, v in self._adapter:
            if k == self._METADATA_KEY:
                continue
            yield v



import redis

from rltk.io.serializer import Serializer, PickleSerializer
from rltk.io.adapter.key_set_adapter import KeySetAdapter


class RedisKeySetAdapter(KeySetAdapter):
    """
    Redis key set adapter.
    
    Args:
        host (str): Host address.
        serializer (Serializer, optional): The serializer used to serialize Record object. 
                                If it's None, `PickleSerializer` will be used. Defaults to None.
        key_prefix (str, optional): Prefix of key in redis. Defaults to empty string.
        clean (bool, optional): Clean adapters while starting. Defaults to False.
        **kwargs: Other parameters used by `redis.Redis <https://redis-py.readthedocs.io/en/latest/#redis.Redis>`_ .
    """

    def __init__(self, host, key_prefix: str = '', serializer: Serializer=None, clean: bool = False, **kwargs):
        if not serializer:
            serializer = PickleSerializer()
        self._redis = redis.Redis(host=host, **kwargs)
        self._serializer = serializer
        self._key_prefix = key_prefix

        if clean:
            self.clean()

    def _encode_key(self, key):
        return '{prefix}{key}'.format(prefix=self._key_prefix, key=key)

    def _decode_key(self, key):
        key = key.decode('utf-8')
        return key[len(self._key_prefix):]

    def get(self, key):
        return self._get(self._encode_key(key))

    def _get(self, key):
        v = set([self._serializer.loads(v) for v in self._redis.smembers(key)])
        if len(v) != 0:
            return v

    def set(self, key, value):
        if not isinstance(value, set):
            raise ValueError('value must be a set')
        self.delete(key)
        for v in value:
            self.add(key, v)

    def add(self, key, value):
        return self._redis.sadd(self._encode_key(key), self._serializer.dumps(value))

    def remove(self, key, value):
        return self._redis.srem(self._encode_key(key), self._serializer.dumps(value))

    def delete(self, key):
        return self._redis.delete(self._encode_key(key))

    def __next__(self):
        # scan_iter() returns generator, keys() returns array
        for key in self._redis.scan_iter(self._encode_key('*')):
            yield self._decode_key(key), self._get(key)



import json
import io


class KeySetAdapter(object):
    """
    Key Set Adapter.
    """

    def get(self, key: str):
        """
        Get value by key.
        
        Args:
            key (str): Key.
            
        Returns:
            set: A set of values, None if key doesn't exist.
        """
        raise NotImplementedError

    def set(self, key: str, value: set):
        """
        Set a set by key.
        
        Args:
            key (str): Key.
            value (builtins.set): Value set.
        """
        raise NotImplementedError

    def add(self, key: str, value: object):
        """
        Add value to a set by key. If key doesn't exist, create one.
        
        Args:
            key (str): Key.
            value (object): Value.
        """
        raise NotImplementedError

    def remove(self, key: str, value: object):
        """
        Remove value from a set by key. If key doesn't exist, create one.
        
        Args:
            key (str): Key.
            value (object): Value.
        """
        raise NotImplementedError

    def delete(self, key: str):
        """
        Delete a set by key.
        
        Args:
            key (str): Key.
        """
        raise NotImplementedError

    def dump(self, f: io.IOBase):
        """
        Dump data to json lines format. Each json object is formatted as `{key: [value1, value2, ...]}`.
        
        Args:
            f (io.IOBase): IO handler.
        """
        for k, ss in self:
            obj = {k: list(ss)}
            f.write(json.dumps(obj) + '\n')

    def clean(self):
        """
        Delete all keys in this adapter.
        """
        for k, _ in self:
            self.delete(k)

    def __init__(self):
        pass

    def __del__(self):
        """
        Same as :meth:`close`.
        """
        self.close()

    def __iter__(self):
        """
        Same as :meth:`__next__`.
        """
        return self.__next__()

    def __next__(self):
        """
        Iterator of the data store. This is not required.

        Returns:
            iter: key, set
        """
        pass

    def close(self):
        """
        Close handler if needed.
        """
        pass



from distutils.core import setup
from setuptools import find_packages

with open('VERSION', 'r') as f:
    version = f.readline().strip()

with open('README.rst', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    install_requires = list()
    dependency_links = list()
    for line in f:
        re = line.strip()
        if re:
            if re.startswith('git+') or re.startswith('svn+') or re.startswith('hg+'):
                dependency_links.append(re)
            else:
                install_requires.append(re)

packages = find_packages()

setup(
    name='rltk',
    version=version,
    packages=packages,
    url='https://github.com/usc-isi-i2/rltk',
    project_urls={
        "Bug Tracker": "https://github.com/usc-isi-i2/rltk/issues",
        "Documentation": "https://rltk.readthedocs.io",
        "Source Code": "https://github.com/usc-isi-i2/rltk",
    },
    license='MIT',
    author='USC/ISI',
    author_email='yixiangy@isi.edu',
    description='Record Linkage ToolKit',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    install_requires=install_requires,
    dependency_links=dependency_links,
    classifiers=(
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology"
    )
)



import csv

from rltk.io.writer import Writer
from rltk.io.io_utils import get_file_handler


class GroundTruthWriter(Writer):
    """
    Ground truth writer.
    
    Args:
        file_handler (io.IOBase): It can be file name or file handler.
    """
    def __init__(self, file_handler):
        self._file_handler = get_file_handler(file_handler, 'w')
        fieldnames = ['id1', 'id2', 'label']
        self._csv_writer = csv.DictWriter(self._file_handler, fieldnames=fieldnames)
        self._csv_writer.writeheader()

    def write(self, id1: str, id2: str, label: bool):
        """
        Writer to file.
        
        Args:
            id1 (str): Id 1.
            id2 (str): Id 2.
            label (bool): Positive (True) or negative (False).
        """
        self._csv_writer.writerow({'id1': id1, 'id2': id2, 'label': label})

    def close(self):
        try:
            self._file_handler.close()
        except:
            pass



from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from rltk.dataset import Dataset
from rltk.blocking.block_generator import BlockGenerator
from rltk.blocking.block import Block
from rltk.blocking.block_black_list import BlockBlackList


class HashBlockGenerator(BlockGenerator):
    """
    Hash block generator.
    """

    def block(self, dataset, function_: Callable = None, property_: str = None,
              block: Block = None, block_black_list: BlockBlackList = None):
        """
        The return of `property_` or `function_` should be string.
        """
        block = super()._block_args_check(function_, property_, block)
        for r in dataset:
            value = function_(r) if function_ else getattr(r, property_)
            if block_black_list and block_black_list.has(value):
                continue
            if not isinstance(value, str):
                raise ValueError('Return of the function or property should be a string')
            block.add(value, dataset.id, r.id)
            if block_black_list:
                block_black_list.add(value, block)
        return block

    def generate(self, block1: Block, block2: Block, output_block: Block = None):
        output_block = super()._generate_args_check(output_block)
        for block_id, ds_id, record_id in block1:
                output_block.add(block_id, ds_id, record_id)
        for block_id, ds_id, record_id in block2:
                output_block.add(block_id, ds_id, record_id)
        return output_block



import rltk.utils as utils


def metaphone(s):
    """
    Metaphone fundamentally improves on the Soundex algorithm by using information about variations and inconsistencies in English spelling and pronunciation to produce a more accurate encoding, which does a better job of matching words and names which sound similar. As with Soundex, similar-sounding words should share the same keys. Metaphone is available as a built-in operator in a number of systems.

    Args:
        s (str): Sequence.

    Returns:
        str: Coded sequence.

    Examples:
        >>> rltk.metaphone('ashcraft')
        'AXKRFT'
        >>> rltk.metaphone('pineapple')
        'PNPL'
    """
    # code from https://github.com/jamesturk/jellyfish
    # Copyright (c) 2015, James Turk
    # Copyright (c) 2015, Sunlight Foundation
    # All rights reserved.

    utils.check_for_none(s)
    utils.check_for_type(str, s)

    s = utils.unicode_normalize(s)

    if len(s) == 0:
        raise ValueError('Empty string')

    s = s.lower()
    result = []

    # skip first character if s starts with these
    if s.startswith(('kn', 'gn', 'pn', 'ac', 'wr', 'ae')):
        s = s[1:]

    i = 0

    while i < len(s):
        c = s[i]
        next_ = s[i+1] if i < len(s)-1 else '*****'
        nextnext = s[i+2] if i < len(s)-2 else '*****'

        # skip doubles except for cc
        if c == next_ and c != 'c':
            i += 1
            continue

        if c in 'aeiou':
            if i == 0 or s[i-1] == ' ':
                result.append(c)
        elif c == 'b':
            if (not (i != 0 and s[i-1] == 'm')) or next_:
                result.append('b')
        elif c == 'c':
            if next_ == 'i' and nextnext == 'a' or next_ == 'h':
                result.append('x')
                i += 1
            elif next_ in 'iey':
                result.append('s')
                i += 1
            else:
                result.append('k')
        elif c == 'd':
            if next_ == 'g' and nextnext in 'iey':
                result.append('j')
                i += 2
            else:
                result.append('t')
        elif c in 'fjlmnr':
            result.append(c)
        elif c == 'g':
            if next_ in 'iey':
                result.append('j')
            elif next_ not in 'hn':
                result.append('k')
            elif next_ == 'h' and nextnext and nextnext not in 'aeiou':
                i += 1
        elif c == 'h':
            if i == 0 or next_ in 'aeiou' or s[i-1] not in 'aeiou':
                result.append('h')
        elif c == 'k':
            if i == 0 or s[i-1] != 'c':
                result.append('k')
        elif c == 'p':
            if next_ == 'h':
                result.append('f')
                i += 1
            else:
                result.append('p')
        elif c == 'q':
            result.append('k')
        elif c == 's':
            if next_ == 'h':
                result.append('x')
                i += 1
            elif next_ == 'i' and nextnext in 'oa':
                result.append('x')
                i += 2
            else:
                result.append('s')
        elif c == 't':
            if next_ == 'i' and nextnext in 'oa':
                result.append('x')
            elif next_ == 'h':
                result.append('0')
                i += 1
            elif next_ != 'c' or nextnext != 'h':
                result.append('t')
        elif c == 'v':
            result.append('f')
        elif c == 'w':
            if i == 0 and next_ == 'h':
                i += 1
            if nextnext in 'aeiou' or nextnext == '*****':
                result.append('w')
        elif c == 'x':
            if i == 0:
                if next_ == 'h' or (next_ == 'i' and nextnext in 'oa'):
                    result.append('x')
                else:
                    result.append('s')
            else:
                result.append('k')
                result.append('s')
        elif c == 'y':
            if next_ in 'aeiou':
                result.append('y')
        elif c == 'z':
            result.append('s')
        elif c == ' ':
            if len(result) > 0 and result[-1] != ' ':
                result.append(' ')

        i += 1

    return ''.join(result).upper()


def metaphone_similarity(s1, s2):
    """
    metaphone(s1) == metaphone(s2)

    Args:
        s1 (str): Sequence.
        s2 (str): Sequence.

    Returns:
        float: if metaphone(s1) equals to metaphone(s2)
    """
    return 1 if metaphone(s1) == metaphone(s2) else 0



import pytest

from rltk.record import Record
from rltk.evaluation.ground_truth import GroundTruth
from rltk.evaluation.trial import Trial
from rltk.similarity import *


class ConcreteRecord(Record):
    @property
    def id(self):
        return self.raw_object['id']

    @property
    def data(self):
        return self.raw_object['data']


@pytest.mark.parametrize('ground_truth_list, cal_result_list, min_c, top_k,tp,tn,fp,fn', [
    ([(1, 1, 10, 0, True), (2, 2, 11, 0, True), (3, 1, 12, 1, False), (4, 2, 13, 1, False)],
     [(1, 1, 10, 0, True, 0.5), (2, 2, 11, 0, False, 0.5), (3, 1, 12, 1, True, 0.5), (4, 2, 13, 1, False, 0.5)], 0, 0,
     0.5, 0.5, 0.5,
     0.5),
    # ([(1, 0, True), (2, 0, True), (1, 1, False), (2, 1, False)],
    #  [(1, 0, True, 0.6), (2, 0, False, 0.5), (1, 1, True, 0.5), (2, 1, False, 0.6)], 0, 2, 1.0, 1.0, 0, 0)
])
def test_basic(ground_truth_list, cal_result_list, min_c, top_k, tp, tn, fp, fn):
    # if not isinstance(ground_truth_list, (list)) or not isinstance(cal_result_list, (list)):
    #     with pytest.raises(ValueError):
    # # number_equal(n1, n2)
    # else:
    do_test_trial(ground_truth_list, cal_result_list, min_c, top_k, tp, tn, fp, fn)


def do_test_trial(ground_truth_list, cal_result_list, min_c, top_k, tp, tn, fp, fn):
    gt = GroundTruth()

    for r1_id, r1_d, r2_id, r2_d, p in ground_truth_list:
        raw_object = {'id': r1_id, 'data': r1_d}
        r1 = ConcreteRecord(raw_object)
        raw_object = {'id': r2_id, 'data': r2_d}
        r2 = ConcreteRecord(raw_object)
        gt.add_ground_truth(r1_id, r2_id, p)

    trial = Trial(gt, min_c, top_k)
    for r1_id, r1_d, r2_id, r2_d, p, c in cal_result_list:
        raw_object = {'id': r1_id, 'data': r1_d}
        r1 = ConcreteRecord(raw_object)
        raw_object = {'id': r2_id, 'data': r2_d}
        r2 = ConcreteRecord(raw_object)
        trial.add_result(r1, r2, p, c)

    trial.evaluate()

    assert trial.true_positives == tp
    assert trial.true_negatives == tn
    assert trial.false_positives == fp
    assert trial.false_negatives == fn


@pytest.mark.parametrize('ground_truth_list, min_c, top_k, similarity_info, tp, tn, fp, fn', [
    ([('0', '', '10', 'abc', False), ('1', 'abc', '11', 'abc', True), ('2', 'abcd', '12', 'abc', False),
      ('3', 'abd', '13', 'abc', False)],
     0, 0, [('levenshtein_similarity', 0.9), ('string_equal', 0.5)], 1.0, 1.0, 0, 0),
    ([('0', '', '10', 'abc', False), ('1', 'abc', '11', 'abc', True), ('2', 'abcd', '12', 'abc', False),
      ('3', 'abd', '13', 'abc', False)],
     0, 2, [('levenshtein_similarity', 0.9), ('string_equal', 0.5)], 1.0, 1.0, 0, 0)
])
def test_lvl(ground_truth_list, min_c, top_k, similarity_info, tp, tn, fp, fn):
    gt = GroundTruth()

    for r1_id, r1_d, r2_id, r2_d, p in ground_truth_list:
        raw_object = {'id': r1_id, 'data': r1_d}
        r1 = ConcreteRecord(raw_object)
        raw_object = {'id': r2_id, 'data': r2_d}
        r2 = ConcreteRecord(raw_object)
        gt.add_ground_truth(r1_id, r2_id, p)

    for similarity_function, min_confidence in similarity_info:
        trial = Trial(gt, min_confidence=min_c, top_k=top_k)

        i = 0
        for r1_id, r1_d, r2_id, r2_d, c in ground_truth_list:
            raw_object = {'id': r1_id, 'data': r1_d}
            r1 = ConcreteRecord(raw_object)
            raw_object = {'id': r2_id, 'data': r2_d}
            r2 = ConcreteRecord(raw_object)

            func_info = similarity_function + '("' + r1_d + '","' + r2_d + '")'
            c = eval(func_info)
            p = (c >= min_confidence)
            trial.add_result(r1, r2, p, c)

        trial.evaluate()

        assert trial.true_positives == tp
        assert trial.true_negatives == tn
        assert trial.false_positives == fp
        assert trial.false_negatives == fn



from rltk.io.reader import Reader


class ArrayReader(Reader):
    """
    Array Reader.
    
    Args:
        array (list): Array.
    """

    def __init__(self, array):
        try:
            for _ in array:
                break
        except TypeError:
            raise TypeError('Can not iterate on ArrayReader')

        self._array = array

    def __next__(self):
        for item in self._array:
            yield item




import sys
import os
import tempfile
import logging

from distributed.cli import dask_scheduler, dask_worker


def help_info():
    print('Available commands:')
    print('remote.worker, remote.scheduler')


if __name__ == '__main__':
    if len(sys.argv) <= 1:
        print('No command\n')
        help_info()
        sys.exit()

    cmd = sys.argv[1]
    sub_cmd = sys.argv[2:] if len(sys.argv) >= 3 else []
    sys.argv.pop(1)

    if cmd in ('help', '--help', 'h', '-h'):
        help_info()
        sys.exit()

    sys.argv[0] = cmd  # replace prog name
    temp_path = os.path.join(tempfile.gettempdir(), 'rltk', 'remote')
    if not os.path.exists(temp_path):
        os.makedirs(temp_path, exist_ok=True)
    if cmd == 'remote.worker':
        logger = logging.getLogger('distributed.dask_worker')
        logger.setLevel(logging.ERROR)
        sys.argv.append('--local-directory')
        sys.argv.append(temp_path)
        # sys.argv.append('--change-directory')
        sys.exit(dask_worker.go())
    elif cmd == 'remote.scheduler':
        logger = logging.getLogger('distributed.scheduler')
        logger.setLevel(logging.ERROR)
        sys.argv.append('--local-directory')
        sys.argv.append(temp_path)
        sys.exit(dask_scheduler.go())
    else:
        print('Unknown command\n')
        help_info()

    sys.exit()



import os
import glob
from distributed import Worker

from dask.distributed import Client
from distributed.security import Security


class Remote(object):
    """
    Remote.
    
    Args:
        address (str): Remote scheduler address formed by `ip:port`.
        tls_ca_file (str, optional): TLS CA certificate file path. Defaults to None.
        tls_client_cert (str, optional): TLS certificate file path. Defaults to None.
        tls_client_key (str, optional): TLS private key file path. Defaults to None.
        require_encryption (bool, optional): Encrypt data exchange. Defaults to False.
        
    Note:
        TLS will be enabled only if all three TLS arguments are provided. 
        Remember to change network protocol to `tls://<address>`.
    """
    def __init__(self, address: str,
                 tls_ca_file: str = None, tls_client_cert: str = None, tls_client_key: str = None,
                 require_encryption: bool = False):
        # authentication
        sec = None
        if tls_ca_file and tls_client_cert and tls_client_key:
            sec = Security(tls_ca_file=tls_ca_file,
                           tls_client_cert=tls_client_cert,
                           tls_client_key=tls_client_key,
                           require_encryption=require_encryption)

        # init
        self._client = Client(address=address, security=sec)
        self._client.register_worker_callbacks(Remote._worker_startup)

    @staticmethod
    def _worker_startup(dask_worker: Worker):
        os.chdir(dask_worker.local_dir)

    def add_dependencies(self, files):
        """
        Add list of dependencies, order matters.
        
        Args:
            files (list): List of dependent files.
        """
        # TODO: automatically resolve module dependencies
        if isinstance(files, str):
            files = [files]
        for f in files:
            self._client.upload_file(f)

    def scatter(self, *args, **kwargs):
        """
        Scatter data.
        """
        return self._client.scatter(*args, **kwargs)

    def submit(self, func, *args, **kwargs):
        """
        Submit function and data.
        
        Args:
            func (callable): User function.
        """
        return self._client.submit(func, *args, **kwargs)

    def fetch(self, futures_, **kwargs):
        """
        Fetch data of future objects.
        
        Args:
            futures_ (list): Future objects.
        """
        return self._client.gather(futures_, **kwargs)

    def cancel(self, futures_, **kwargs):
        """
        Cancel job of future objects.
        
        Args:
            futures_ (list): Future objects.
        """
        return self._client.cancel(futures_, **kwargs)

    def close(self, *args, **kwargs):
        """
        Close connection.
        """
        return self._client.close(*args, **kwargs)

    # @staticmethod
    # def _list_local_dir(pathname='**', *args, recursive=True):
    #     non_py_files = []
    #     py_files = []
    #     for path in glob.glob(pathname, *args, recursive=recursive):
    #         if os.path.isdir(path):
    #             if path == '__pycache__':
    #                 continue
    #         elif os.path.isfile(path):
    #             if path.endswith('.pyc'):
    #                 continue
    #             if path.endswith('.py'):
    #                 py_files.append(path)
    #             else:
    #                 non_py_files.append(path)
    #
    #     return non_py_files + py_files



import redis

from rltk.record import Record
from rltk.io.adapter import KeyValueAdapter
from rltk.io.serializer import Serializer, PickleSerializer


class RedisKeyValueAdapter(KeyValueAdapter):
    """
    Redis adapter.
    
    Args:
        host (str): Host address.
        serializer (Serializer, optional): The serializer used to serialize Record object. 
                                If it's None, `PickleSerializer` will be used. Defaults to None.
        key_prefix (str, optional): Prefix of key in redis. Defaults to empty string. 
        clean (bool, optional): Clean adapters while starting. Defaults to False.
        **kwargs: Other parameters used by `redis.Redis <https://redis-py.readthedocs.io/en/latest/#redis.Redis>`_ .
    """
    def __init__(self, host, serializer: Serializer=None, key_prefix: str = '', clean: bool = False, **kwargs):
        if not serializer:
            serializer = PickleSerializer()
        self._redis = redis.Redis(host=host, **kwargs)
        self._serializer = serializer
        self._key_prefix = key_prefix

        if clean:
            self.clean()

    #: parallel-safe
    parallel_safe = True

    def _encode_key(self, key):
        return self._key_prefix + key

    def _decode_key(self, key):
        key = key.decode('utf-8')
        return key[len(self._key_prefix):]

    def get(self, key) -> object:
        v = self._redis.get(self._encode_key(key))
        if v:
            return self._serializer.loads(v)

    def set(self, key, value: object):
        return self._redis.set(self._encode_key(key), self._serializer.dumps(value))

    def delete(self, key):
        return self._redis.delete(self._encode_key(key))

    def __next__(self):
        # scan_iter() returns generator, keys() returns array
        for key in self._redis.scan_iter(self._encode_key('*')):
            yield self._decode_key(key), self._serializer.loads(self._redis.get(key))



import happybase

from rltk.record import Record
from rltk.io.adapter import KeyValueAdapter
from rltk.io.serializer import Serializer, PickleSerializer


class HBaseKeyValueAdapter(KeyValueAdapter):
    """
    Hbase Adapter.
    
    Args:
        host (str): Host address.
        table (str): HBase table name.
        serializer (Serializer, optional): The serializer used to serialize Record object. 
                                If it's None, `PickleSerializer` will be used. Defaults to None.
        key_prefix (str, optional): The prefix of HBase row key.
        clean (bool, optional): Clean adapters while starting. Defaults to False.
        **kwargs: Other parameters used by `happybase.Connection <https://happybase.readthedocs.io/en/latest/api.html#connection>`_ .
    
    Note:
        The timeout of thrift in hbase-site.xml needs to increase::
        
            <property>
                <name>hbase.thrift.server.socket.read.timeout</name>
                <value>6000000</value>
            </property>
            <property>
                <name>hbase.thrift.connection.max-idletime</name>
                <value>18000000</value>
            </property>
    """

    def __init__(self, host, table, serializer: Serializer = None, key_prefix: str = '', clean: bool = False, **kwargs):
        if not serializer:
            serializer = PickleSerializer()
        self._conn = happybase.Connection(host=host, timeout=None, **kwargs)
        self._serializer = serializer
        self._key_prefix = key_prefix
        self._family_name = 'rltk'
        self._col_name = 'obj'
        self._fam_col_name = '{}:{}'.format(self._family_name, self._col_name).encode('utf-8')

        if table.encode('utf-8') not in self._conn.tables():
            self._create_table(table)
        self._table = self._conn.table(table)

        if clean:
            self.clean()

    #: parallel-safe
    parallel_safe = True

    def _encode_key(self, key):
        return '{prefix}{key}'.format(prefix=self._key_prefix, key=key).encode('utf-8')

    def _decode_key(self, key):
        key = key.decode('utf-8')
        return key[len(self._key_prefix):]

    def close(self):
        try:
            self._conn.close()
        except:
            pass

    def _create_table(self, table_name):
        self._conn.create_table(table_name, {self._family_name: dict()})

    def get(self, key) -> object:
        return self._serializer.loads(self._table.row(self._encode_key(key))[self._fam_col_name])

    def set(self, key, value: object):
        return self._table.put(self._encode_key(key), {self._fam_col_name: self._serializer.dumps(value)})

    def delete(self, key):
        return self._table.delete(self._encode_key(key))

    def __next__(self):
        for key, data in self._table.scan(
                row_prefix=self._key_prefix.encode('utf-8'), filter=b'FirstKeyOnlyFilter()'):
            yield self._decode_key(key), self._serializer.loads(data[self._fam_col_name])



"""
This module is designed for breaking the restriction of Python Global Interpreter Lock (GIL): It uses multi-processing (compute-intensive operations) and multi-threading (return data collecting) to accelerate computing.
Once it's initialized, it creates a sub process pool, all the added data will be dispatched to different sub processes for parallel computing. The result sends back and consumes in another thread in current main process. The Inter Process Communication (IPC) between main process and sub processes is based on queue.

Example::

    result = []
    
    def dummy_computation_with_input(x):
        time.sleep(0.0001)
        return x * x, x + 5
    
    def output_handler(r1, r2):
        result.append(r1 if r1 > r2 else r2)
    
    pp = ParallelProcessor(dummy_computation_with_input, 8, output_handler=output_handler)
    pp.start()
    
    for i in range(8):
        pp.compute(i)
    
    pp.task_done()
    pp.join()
    
    print(result)
"""

import multiprocessing as mp
import threading
import queue
from typing import Callable


class OutputThread(threading.Thread):
    """
    Handle output in main process.
    Create a thread and call ParallelProcessor.get_output().
    """
    def __init__(self, instance, output_handler):
        super(OutputThread, self).__init__()
        self.output_handler = output_handler
        self.instance = instance

    def run(self):
        for batched_output in self.instance.get_output():
            for o in batched_output:
                self.output_handler(*o)


class ParallelProcessor(object):
    """
    Args:
        input_handler (Callable): Computational function. 
        num_of_processor (int): Number of processes to use.
        max_size_per_input_queue (int, optional): Maximum size of input queue for one process.
                                    If it's full, the corresponding process will be blocked.
                                    0 by default means unlimited.
        max_size_per_output_queue (int, optional): Maximum size of output queue for one process.
                                    If it's full, the corresponding process will be blocked.
                                    0 by default means unlimited.
        output_handler (Callable, optional): If the output data needs to be get in main process (another thread),
                                set this handler, the arguments are same to the return from input_handler.
                                The return result is one by one, order is arbitrary.
        enable_process_id (bool, optional): If it's true, an additional argument `_idx` (process id) will be
                                passed to `input_handler`. It defaults to False.
        batch_size (int, optional): Batch size, defaults to 1.
    
    
    Note:
        Do NOT implement heavy compute-intensive operations in output_handler, they should be in input_handler.
    """

    # Command format in queue. Represent in tuple.
    # The first element of tuple will be command, the rests are arguments or data.
    # (CMD_XXX, args...)
    CMD_DATA = 0
    CMD_STOP = 1

    def __init__(self, input_handler: Callable, num_of_processor: int,
                 max_size_per_input_queue: int = 0, max_size_per_output_queue: int = 0,
                 output_handler: Callable = None, enable_process_id: bool = False,
                 batch_size: int = 1):
        self.num_of_processor = num_of_processor
        self.input_queues = [mp.Queue(maxsize=max_size_per_input_queue) for _ in range(num_of_processor)]
        self.output_queues = [mp.Queue(maxsize=max_size_per_output_queue) for _ in range(num_of_processor)]
        self.processes = [mp.Process(target=self.run, args=(i, self.input_queues[i], self.output_queues[i]))
                          for i in range(num_of_processor)]
        self.input_handler = input_handler
        self.output_handler = output_handler
        self.input_queue_index = 0
        self.output_queue_index = 0
        self.enable_process_id = enable_process_id
        self.batch_size = batch_size
        self.batch_data = []

        # output can be handled in each process or in main process after merging (output_handler needs to be set)
        # if output_handler is set, output needs to be handled in main process; otherwise, it assumes there's no output.
        if output_handler:
            self.output_thread = OutputThread(self, output_handler)

    def start(self):
        """
        Start processes and threads.
        """
        if self.output_handler:
            self.output_thread.start()
        for p in self.processes:
            p.start()

    def join(self):
        """
        Block until processes and threads return.
        """
        for p in self.processes:
            p.join()
        if self.output_handler:
            self.output_thread.join()

    def task_done(self):
        """
        Indicate that all resources which need to compute are added to processes.
        (main process, blocked)
        """
        if len(self.batch_data) > 0:
            self._compute(self.batch_data)
            self.batch_data = []

        for q in self.input_queues:
            q.put( (ParallelProcessor.CMD_STOP,) )

    def compute(self, *args, **kwargs):
        """
        Add data to one of the input queues.
        (main process, unblocked, using round robin to find next available queue)
        """
        self.batch_data.append( (args, kwargs) )

        if len(self.batch_data) == self.batch_size:
            self._compute(self.batch_data)
            self.batch_data = []  # reset buffer

    def _compute(self, batched_args):
        while True:
            q = self.input_queues[self.input_queue_index]
            self.input_queue_index = (self.input_queue_index + 1) % self.num_of_processor
            try:
                q.put_nowait( (ParallelProcessor.CMD_DATA, batched_args) )
                return  # put in
            except queue.Full:
                continue  # find next available

    def run(self, idx: int, input_queue: mp.Queue, output_queue: mp.Queue):
        """
        Process’s activity. It handles queue IO and invokes user's input handler.
        (subprocess, blocked, only two queues can be used to communicate with main process)
        """
        while True:
            data = input_queue.get()
            if data[0] == ParallelProcessor.CMD_STOP:
                # print(idx, 'stop')
                if self.output_handler:
                    output_queue.put( (ParallelProcessor.CMD_STOP,) )
                return
            elif data[0] == ParallelProcessor.CMD_DATA:
                batch_result = []
                for d in data[1]:
                    args, kwargs = d[0], d[1]
                    # print(idx, 'data')
                    result = self.input_handler(*args, **kwargs, _idx=idx) if self.enable_process_id \
                        else self.input_handler(*args, **kwargs)
                    if self.output_handler:
                        if not isinstance(result, tuple):  # output must represent as tuple
                            result = (result,)
                        batch_result.append(result)
                if len(batch_result) > 0:
                    output_queue.put( (ParallelProcessor.CMD_DATA, batch_result) )
                    batch_result = []  # reset buffer

    def get_output(self):
        """
        Get data from output queue sequentially.
        (main process, unblocked, using round robin to find next available queue)
        """
        if not self.output_handler:
            return
        while True:
            # print(self.output_queues)
            q = self.output_queues[self.output_queue_index]
            try:
                data = q.get_nowait()  # get out
                if data[0] == ParallelProcessor.CMD_STOP:
                    del self.output_queues[self.output_queue_index]  # remove queue if it's finished
                elif data[0] == ParallelProcessor.CMD_DATA:
                    yield data[1]
            except queue.Empty:
                continue  # find next available
            finally:
                if len(self.output_queues) == 0:  # all finished
                    return
                self.output_queue_index = (self.output_queue_index + 1) % len(self.output_queues)



import io


def get_file_handler(f, mode='r'):
    """
    Helper function for getting file handler.

    Args:
        f (Union[str,io.IOBase]): File path or handler.
        mode (str, optional): Parameter mode in :py:meth:`open`. Defaults to `r`.

    Returns:
        io.IOBase: File handler.
    """
    if isinstance(f, io.IOBase):
        return f

    return open(f, mode)
